from pathlib import Path
import attrs
import json
from fileformats.core import mark
from fileformats.core.utils import MissingExtendedDependency
from fileformats.medimage.base import MedicalImage
from fileformats.medimage.dicom import Dicom
from fileformats.medimage import (
    Analyze,
    Nifti,
    NiftiGz,
    NiftiX,
    NiftiGzX,
    MrtrixImage,
    MrtrixImageHeader,
    NiftiXBvec,
    NiftiBvec,
    NiftiGzBvec,
    NiftiGzXBvec,
)
try:
    import jq
except ImportError:
    jq = MissingExtendedDependency("jq", __name__)
try:
    import pydra
except ImportError:
    pydra = MissingExtendedDependency("pydra", __name__)
try:
    import pydra.tasks.mrtrix3.utils as pydra_mrtrix3_utils
except ImportError:
    pydra_mrtrix3_utils = MissingExtendedDependency("pydra.tasks.mrtrix3", __name__)
try:
    import pydra.tasks.dcm2niix as pydra_dcm2niix
except ImportError:
    pydra_dcm2niix = MissingExtendedDependency("pydra.tasks.dcm2niix", __name__)


@mark.converter(source_format=MedicalImage, target_format=Analyze, out_ext=Analyze.ext)
@mark.converter(
    source_format=MedicalImage, target_format=MrtrixImage, out_ext=MrtrixImage.ext
)
@mark.converter(
    source_format=MedicalImage,
    target_format=MrtrixImageHeader,
    out_ext=MrtrixImageHeader.ext,
)
def mrconvert(name, out_ext: str):
    """Initiate an MRConvert task with the output file extension set

    Parameters
    ----------
    name : str
        name of the converter task
    out_ext : str
        extension of the output file, used by MRConvert to determine the desired format

    Returns
    -------
    pydra.ShellCommandTask
        the converter task
    """
    return pydra_mrtrix3_utils.MRConvert(name=name, out_file="out" + out_ext)


@mark.converter(source_format=Dicom, target_format=Nifti)
@mark.converter(source_format=Dicom, target_format=NiftiGz, compress="y")
@mark.converter(source_format=Dicom, target_format=NiftiX)
@mark.converter(source_format=Dicom, target_format=NiftiGzX, compress="y")
@mark.converter(source_format=Dicom, target_format=NiftiXBvec)
@mark.converter(source_format=Dicom, target_format=NiftiBvec)
@mark.converter(source_format=Dicom, target_format=NiftiGzBvec)
@mark.converter(
    source_format=Dicom, target_format=NiftiGzXBvec, compress="y"
)
def extended_dcm2niix(
    name,
    compress: str = "n",
    file_postfix: str = None,
    side_car_jq: str = None,
    extract_volume: int = None,
    to_4d: bool = False,
):
    """The Dcm2niix command wrapped in a workflow in order to map the inputs and outputs
    onto "in_file" and "out_file", respectively, and implement optional post-conversion
    manipulations to allow manual override of conversion issues.

    Parameters
    ----------
    name : str
        name of the workflow
    compress : str, optional
        whether to apply compression to the conversion, by default "n"
    file_postfix : str, optional
        select one of the multiple output files by its generated postfix.
        See https://github.com/rordenlab/dcm2niix/blob/master/FILENAMING.md for
        complete list of different postfixes that will be generated, by default None
    side_car_jq : str, optional
        JQ (https://stedolan.github.io/jq/) expression that can be provided to edit
        the resulting JSON side-car. Can be used to fix any conversion issues or add
        required fields manually, by default None
    extract_volume : int, optional
        extract a 3D volume from a 4D dataset by passing the index (0-based) of the
        volume to extract, by default None
    to_4d : bool, optional
        whether to wrap resulting 3D NIfTI volume in a 4D dataset, by default False

    Returns
    -------
    pydra.Workflow
        the converter workflow

    Raises
    ------
    ValueError
        when mutually exclusive "extract_volume" and "to_4d" options are provided
    """

    if extract_volume is not None and to_4d:
        raise ValueError(
            f"'extract_volume' ({extract_volume}) and 'to_4d' are mutually exclusive"
        )
    # Create workflow to map input field to "in_file" and optionally perform post-conversion
    # steps to manipulate the converted NIfTI files
    wf = pydra.Workflow(
        name=name,
        input_spec=["in_file"],
    )
    if file_postfix is None:
        file_postfix = attrs.NOTHING
    wf.add(
        pydra_dcm2niix.Dcm2Niix(
            in_dir=wf.lzin.in_file,
            out_dir=".",
            name="dcm2niix",
            compress=compress,
            file_postfix=file_postfix,
        )
    )
    out_file = wf.dcm2niix.lzout.out_file
    out_json = wf.dcm2niix.lzout.out_json
    # Add MRConvert step to either select a single volume of a 4D dataset or the inverse,
    # wrap a single volume in a 4D dataset
    if extract_volume is not None or to_4d:
        if extract_volume:
            coord = [3, extract_volume]
            axes = [0, 1, 2]
        else:  # to_4d
            coord = attrs.NOTHING
            axes = [0, 1, 2, -1]
        wf.add(
            pydra_mrtrix3_utils.MRConvert(
                in_file=out_file,
                coord=coord,
                axes=axes,
                name="mrconvert",
            )
        )
        out_file = wf.mrconvert.lzout.out_file
    # Add JQ edit of side car to allow manual fixing of any conversion issues
    if side_car_jq is not None:
        wf.add(
            edit_dcm2niix_side_car(
                in_file=out_json, jq_expr=side_car_jq, name="json_edit"
            )
        )
        out_json = wf.json_edit.lzout.out

    wf.add(
        collect_dcm2niix_outputs(
            name="collect_outputs",
            out_file=out_file,
            out_json=out_json,
            out_bvec=wf.dcm2niix.lzout.out_bvec,
            out_bval=wf.dcm2niix.lzout.out_bval,
        )
    )
    # Set workflow outputs
    wf.set_output(("out_file", wf.collect_outputs.lzout.out))
    return wf


@pydra.mark.task
def edit_dcm2niix_side_car(in_file: Path, jq_expr: str, out_file=None) -> Path:
    """ "Applies ad-hoc edit of JSON side car with JQ query language"""
    if out_file is None:
        out_file = in_file
    with open(in_file) as f:
        dct = json.load(f)
    dct = jq.compile(jq_expr).input(dct).first()
    with open(out_file, "w") as f:
        json.dump(dct, f)
    return in_file


@pydra.mark.task
def collect_dcm2niix_outputs(
    out_file: Path, out_json: Path, out_bvec: Path, out_bval: Path
) -> list[Path]:
    lst = [out_file]
    for file in (out_json, out_bvec, out_bval):
        if file is not attrs.NOTHING:
            lst.append(file)
    return lst



from fileformats.generic import File


class ListMode(File):

    ext = ".bf"
    binary = True


class Kspace(File):

    binary = True
    iana_mime = None


class TwixVb(Kspace):
    """The format that k-space data is saved in from Siemens scanners
    with system version vB to (at least) vE"""

    ext = ".dat"


# class CustomKspace(Kspace):
#     """A custom format for saving k-space data in binary amd JSON files.

#     Binary files
#     ------------
#     primary : 5-d matrix
#         Data from "data" scan organised in the following dimension order:
#         channel, freq-encode, phase-encode, partition-encode (slice), echoes
#     reference : 5-d matrix
#         Data from calibration scan organised in the same dimension order as
#         primary scan

#     JSON side-car
#     -------------
#     dims : 3-tuple(int)
#         The dimensions of the image in freq, phase, partition (slice) order
#     voxel_size : 3-tuple(float)
#         Size of the voxels in same order as dims
#     num_channels : int
#         Number of channels in the k-space
#     num_echos : int
#         Number of echoes in the acquisition
#     T E : tuple(float)
#         The echo times
#     B0_strength : float
#         Strength of the B0 field
#     B0_dir : 3-tuple(float)
#         Direction of the B0 field
#     larmor_freq : float
#         The central larmor row_frequency of the scanner"""

#     ext = ".ks"
#     side_cars = ("ref", "json")


class Rda(File):
    """MRS format"""

    ext = ".rda"
    binary = True



import numpy as np
from fileformats.core import mark
from fileformats.core.mixin import WithAdjacentFiles
from fileformats.generic import File
from .nifti import NiftiGzX, NiftiGz, Nifti1, NiftiX


class DwiEncoding(File):

    iana_mime = None

    @property
    def array(self):
        "Both the gradient direction and weighting combined into a single Nx4 array"
        raise NotImplementedError(
            f"array property hasn't been implemented for {type(self)} diffusion encoding subclass"
        )

    @property
    def directions(self):
        "Both the gradient direction and weighting combined into a single Nx4 array"
        raise NotImplementedError(
            f"array property hasn't been implemented for {type(self)} diffusion encoding subclass"
        )

    @property
    def b_values(self):
        raise NotImplementedError(
            f"array property hasn't been implemented for {type(self)} diffusion encoding subclass"
        )


class Bval(File):

    ext = ".bval"

    @property
    def array(self):
        return np.asarray([float(ln) for ln in self.read_contents().split()])


class Bvec(WithAdjacentFiles, DwiEncoding):
    """FSL-style diffusion encoding, in two separate files"""

    ext = ".bvec"
    header_type = Bval

    @property
    def array(self) -> np.ndarray:
        return np.concatenate(self.directions, self.b_values, axis=1)

    @mark.required
    @property
    def b_values_file(self) -> Bval:
        return Bval(self.select_by_ext(Bval))

    @property
    def directions(self) -> np.ndarray:
        return np.asarray(
            [[float(x) for x in ln.split()] for ln in self.read_contents().splitlines()]
        ).T

    @property
    def b_values(self):
        return self.b_values_file.array


class Bfile(DwiEncoding):
    """MRtrix-style diffusion encoding, all in one file"""

    ext = ".b"

    @property
    def array(self) -> np.ndarray:
        return np.asarray(
            [[float(x) for x in ln.split()] for ln in self.read_contents().splitlines()]
        )

    @property
    def directions(self) -> np.ndarray:
        return self.array[:, :3]

    @property
    def b_values(self) -> np.ndarray:
        return self.array[:, 3]


# NIfTI file format gzipped with BIDS side car
class WithBvec(WithAdjacentFiles):
    @mark.required
    @property
    def encoding(self) -> Bvec:
        return Bvec(self.select_by_ext(Bvec))


# NIfTI file format gzipped with BIDS side car
class WithBfile(WithAdjacentFiles):
    @mark.required
    @property
    def encoding(self) -> Bfile:
        return Bfile(self.select_by_ext(Bfile))


class NiftiBvec(WithBvec, Nifti1):
    iana_mime = "application/x-nifti2+bvec"


class NiftiGzBvec(WithBvec, NiftiGz):
    iana_mime = "application/x-nifti2+gzip.bvec"


class NiftiXBvec(WithBvec, NiftiX):
    iana_mime = "application/x-nifti2+json.bvec"


class NiftiGzXBvec(WithBvec, NiftiGzX):
    iana_mime = "application/x-nifti2+gzip.json.bvec"


class NiftiB(WithBfile, Nifti1):
    iana_mime = "application/x-nifti2+b"


class NiftiGzB(WithBfile, NiftiGz):
    iana_mime = "application/x-nifti2+gzip.b"


class NiftiXB(WithBfile, NiftiX):
    iana_mime = "application/x-nifti2+json.b"


class NiftiGzXB(WithBfile, NiftiGzX):
    iana_mime = "application/x-nifti2+gzip.json.b"


# Track files


class MrtrixTrack(File):

    ext = ".tck"



from fileformats.generic import File
from fileformats.core import mark
from fileformats.core.mixin import (
    WithSideCars, WithMagicNumber, WithSeparateHeader, WithAdjacentFiles)
from fileformats.serialization import Json
# from fileformats.text import Tsv
from fileformats.archive import Gzip
from .base import NeuroImage


class Nifti(NeuroImage):

    ext = ".nii"

    def load_metadata(self):
        import nibabel
        return dict(nibabel.load(self.fspath).header)

    @property
    def data_array(self):
        import nibabel
        return nibabel.load(self.fspath).get_data()

    @property
    def vox_sizes(self):
        # FIXME: This won't work for 4-D files
        return self.metadata["pixdim"][1:4]

    @property
    def dims(self):
        # FIXME: This won't work for 4-D files
        return self.metadata["dim"][1:4]


class WithBids(WithSideCars):

    primary_type = Nifti
    side_car_types = (Json,)

    @mark.required
    @property
    def json_file(self):
        return Json(self.select_by_ext(Json))

    # @mark.required
    # @property
    # def tsv_file(self):
    #     return Json(self.select_by_ext(Json, allow_none=True))


class Nifti1(WithMagicNumber, Nifti):

    iana_mime = "application/x-nifti1"
    magic_number = "6E2B3100"
    magic_number_offset = 344


class Nifti2(WithMagicNumber, Nifti):

    ext = ".nii"
    iana_mime = "application/x-nifti2"
    magic_number = "6e2b3200"
    magic_number_offset = 344


class NiftiGz(Nifti, Gzip):  # Should be Gzip[Nifti1]

    ext = ".nii.gz"
    iana_mime = "application/x-nifti1+gzip"


class NiftiX(WithBids, Nifti):
    iana_mime = "application/x-nifti1+json"


class NiftiGzX(WithBids, NiftiGz):
    iana_mime = "application/x-nifti1+gzip.bids"


class NiftiDataFile(NeuroImage):

    ext = ".img"


class NiftiWithDataFile(WithAdjacentFiles, Nifti1):

    magic_number = "6E693100"
    alternate_exts = (".hdr",)

    @mark.required
    @property
    def data_file(self):
        return self.select_by_ext(NiftiDataFile)


class AnalyzeHeader(File):

    ext = ".hdr"
    binary = True

    @property
    def load(self):
        raise NotImplementedError


class Analyze(WithSeparateHeader, NeuroImage):

    ext = ".img"
    header_type = AnalyzeHeader

    @property
    def data_array(self):
        raise NotImplementedError



import numpy as np
from fileformats.generic import File


# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class MedicalImage(File):

    iana_mime = None
    INCLUDE_HDR_KEYS = None
    IGNORE_HDR_KEYS = None
    binary = True

    @property
    def data_array(self):
        """
        Returns the binary data of the image in a numpy array
        """
        raise NotImplementedError

    def contents_equal(self, other_image, rms_tol=None, **kwargs):
        """
        Test whether the (relevant) contents of two image self are equal
        given specific criteria

        Parameters
        ----------
        other_image : Fileset
            The other self to compare
        rms_tol : float
            The root-mean-square tolerance that is acceptable between the array
            data for the images to be considered equal
        """
        if type(other_image) != type(self):
            return False
        if self.headers_diff(self, other_image, **kwargs):
            return False
        if rms_tol:
            rms_diff = self.rms_diff(self, other_image)
            return rms_diff < rms_tol
        else:
            return np.array_equiv(self.get_array(), other_image.get_array())

    def metadata_diff(self, other_image, include_keys=None, ignore_keys=None, **kwargs):
        """
        Check headers to see if all values
        """
        diff = []
        hdr = self.get_header()
        hdr_keys = set(hdr.keys())
        other_hdr = other_image.get_header()
        if include_keys is not None:
            if ignore_keys is not None:
                raise ValueError(
                    "Doesn't make sense to provide both 'include_keys' ({}) "
                    "and ignore_keys ({}) to headers_equal method".format(
                        include_keys, ignore_keys
                    )
                )
            include_keys &= hdr_keys
        elif ignore_keys is not None:
            include_keys = hdr_keys - set(ignore_keys)
        else:
            if self.INCLUDE_HDR_KEYS is not None:
                if self.IGNORE_HDR_KEYS is not None:
                    raise ValueError(
                        "Doesn't make sense to have both 'INCLUDE_HDR_FIELDS'"
                        "and 'IGNORE_HDR_FIELDS' class attributes of class {}".format(
                            type(self).__name__
                        )
                    )
                include_keys = self.INCLUDE_HDR_KEYS  # noqa pylint: disable=no-member
            elif self.IGNORE_HDR_KEYS is not None:
                include_keys = hdr_keys - set(self.IGNORE_HDR_KEYS)
            else:
                include_keys = hdr_keys
        for key in include_keys:
            value = hdr[key]
            try:
                other_value = other_hdr[key]
            except KeyError:
                diff.append(key)
            else:
                if isinstance(value, np.ndarray):
                    if not isinstance(other_value, np.ndarray):
                        diff.append(key)
                    else:
                        try:
                            if not np.allclose(value, other_value, equal_nan=True):
                                diff.append(key)
                        except TypeError:
                            # Fallback to a straight comparison for some formats
                            if value != other_value:
                                diff.append(key)
                elif value != other_value:
                    diff.append(key)
        return diff

    def rms_diff(self, other_image):
        """
        Return the RMS difference between the image arrays
        """
        return np.sqrt(np.sum((self.get_array() - other_image.get_array()) ** 2))


class NeuroImage(MedicalImage):
    """Imaging formats developed for neuroimaging scans"""
    iana_mime = None



import os
import os.path as op
import pydicom
import numpy as np
from fileformats.generic import File, BaseDirectory
from .base import MedicalImage

# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class DicomFile(
    File
):  # FIXME: Should extend from MedicalImage, but need to implement header and array

    ext = ".dcm"


class SiemensDicomFile(DicomFile):

    ext = ".IMA"


class Dicom(BaseDirectory, MedicalImage):

    content_types = (DicomFile,)
    alternate_names = ("secondary",)

    SERIES_NUMBER_TAG = ("0020", "0011")

    def dcm_files(self):
        return [f for f in os.listdir(self.path) if f.endswith(".dcm")]

    @property
    def data_array(self):
        image_stack = []
        for fname in self.dcm_files(self):
            image_stack.append(pydicom.dcmread(op.join(self.path, fname)).pixel_array)
        return np.asarray(image_stack)

    def load_metadata(self, index=0):
        dcm_files = [f for f in os.listdir(self.path) if f.endswith(".dcm")]
        # TODO: Probably should collate fields that vary across the set of
        #       files in the set into lists
        return pydicom.dcmread(op.join(self.path, dcm_files[index]))

    @property
    def vox_sizes(self):
        return np.array(self.metadata["PixelSpacing"] + [self.metadata["SliceThickness"]])

    @property
    def dims(self):
        return np.array(
            (self.metadata["Rows"], self.metadata["DataColumns"], len(self.dcm_files(self))), datatype=int
        )

    def extract_id(self):
        return int(self.dicom_values([self.SERIES_NUMBER_TAG])[0])

    def dicom_values(self, tags):
        """
        Returns a dictionary with the DICOM header fields corresponding
        to the given tag names

        Parameters
        ----------
        fileset : FileSet
            The file set to extract the DICOM header for
        tags : List[Tuple[str, str]]
            List of DICOM tag values as 2-tuple of strings, e.g.
            [('0080', '0020')]

        Returns
        -------
        dct : Dict[Tuple[str, str], str|int|float]
        """

        def read_header():
            dcm = self.get_header(0)
            return [dcm[t].value for t in tags]

        try:
            if self.fspath:
                # Get the DICOM object for the first file in the self
                dct = read_header()
            else:
                try:
                    # Try to access dicom header details remotely
                    hdr = self.row.dataset.store.dicom_header(self)
                except AttributeError:
                    self.get()  # Fallback to downloading data to read header
                    dct = read_header()
                else:
                    dct = [hdr[t] for t in tags]
        except KeyError as e:
            e.msg = "{} does not have dicom tag {}".format(self, str(e))
            raise e
        return dct


class SiemensDicom(Dicom):

    content_types = (SiemensDicomFile,)
    alternative_names = ("dicom",)



from fileformats.generic import File
from fileformats.core.mixin import WithSeparateHeader
from .base import NeuroImage


# ==================
# Other Data Formats
# ==================


class AnalyzeHeader(File):

    ext = ".hdr"

    @property
    def load(self):
        raise NotImplementedError


class Analyze(WithSeparateHeader, NeuroImage):

    ext = ".img"
    header_type = AnalyzeHeader

    @property
    def data_array(self):
        raise NotImplementedError
